from typing import Optional

from PySide6 import QtWidgets

from asphodel.device_info import DeviceInfo
from hyperborea.setting_widget import SettingWidget

from .ui.ui_setting_viewer import Ui_SettingViewerDialog


class SettingViewerDialog(Ui_SettingViewerDialog, QtWidgets.QDialog):
    def __init__(self, device_info: DeviceInfo,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        # capture these now
        self.settings = device_info.settings
        self.nvm_bytes = device_info.nvm
        self.custom_enums = device_info.custom_enums
        self.setting_categories = device_info.setting_categories

        self.setting_widgets: list[SettingWidget] = []

        self.setupUi(self)

        self.add_setting_widgets()

    def add_setting_widgets(self) -> None:
        remaining_settings = set(range(len(self.settings)))

        setting_tabs: list[tuple[str,
                                 list[tuple[QtWidgets.QWidget, ...]]]] = []

        for name, category_settings in self.setting_categories:
            widgets: list[tuple[QtWidgets.QWidget, ...]] = []
            for setting_index in category_settings:
                if setting_index in remaining_settings:
                    remaining_settings.remove(setting_index)

                    setting = self.settings[setting_index]
                    widget = SettingWidget(setting, self.nvm_bytes,
                                           self.custom_enums)
                    self.setting_widgets.append(widget)
                    widgets.append(widget.widgets)
            setting_tabs.append((name, widgets))

        if remaining_settings:
            remaining_widgets: list[tuple[QtWidgets.QWidget, ...]] = []
            for setting_index in sorted(remaining_settings):
                setting = self.settings[setting_index]
                widget = SettingWidget(setting, self.nvm_bytes,
                                       self.custom_enums)
                self.setting_widgets.append(widget)
                remaining_widgets.append(widget.widgets)
            # add the default tab at the beginning
            setting_tabs.insert(0, ("Device Settings", remaining_widgets))

        for tab_name, widgets in setting_tabs:
            tab_widget = QtWidgets.QWidget()
            form_layout = QtWidgets.QFormLayout(tab_widget)

            for row_widgets in widgets:
                form_layout.addRow(*row_widgets)

            self.tabWidget.addTab(tab_widget, tab_name)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from matplotlib.backend_bases import MouseButton
from matplotlib.figure import Figure
import numpy
from numpy.typing import NDArray
from PySide6 import QtGui, QtWidgets

from . import util
from .. import export_script

logger = logging.getLogger(__name__)


def get_mean_std_dev_in_range(
        start_time: float, end_time: float, x: NDArray[numpy.float64],
        y: NDArray[numpy.float64]) -> tuple[float, float]:
    start_index: int = max(0, numpy.searchsorted(
        x, start_time, side="right").item() - 1)
    end_index: int = min(len(x), numpy.searchsorted(
        x, end_time, side="left").item() + 1)

    data = y[start_index:end_index]

    mean: float = numpy.mean(data, axis=0).item()
    std_dev: float = numpy.std(data, axis=0).item()

    return mean, std_dev


def display_mean_and_std_dev(
        start_time: float, end_time: float,
        data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                         str, util.UnitInfo]],
        title: str, parent: QtWidgets.QWidget) -> None:

    value_strings: list[str] = []

    for x, y, name, unit in data:
        mean, std_dev = get_mean_std_dev_in_range(start_time, end_time, x, y)

        base_str = f"mean={mean} {unit.utf8}, std dev={std_dev} {unit.utf8}"

        if name:
            value_strings.append(f"{name}: {base_str}")
        else:
            value_strings.append(base_str)

    text = "\n".join(value_strings)

    QtWidgets.QMessageBox.information(parent, title, text)


def add_mean_std_dev_summary_buttons(fig: Figure, axes: list, data: list[
        tuple[NDArray[numpy.float64], NDArray[numpy.float64], str,
              util.UnitInfo]]) -> None:
    rects = {ax: None for ax in axes}  # used in the callback

    toolbar: QtWidgets.QToolBar = fig.canvas.toolbar  # type: ignore
    total_action_text = QtWidgets.QApplication.translate(
        "RMSTotalAction", "Mean and Std Dev (All)")
    total_action = QtGui.QAction(total_action_text, toolbar)
    total_action.setIcon(QtGui.QIcon.fromTheme("multimeter_analog"))

    range_action_text = QtWidgets.QApplication.translate(
        "RMSRangeAction", "Mean and Std Dev (Range)")
    range_action = QtGui.QAction(range_action_text, toolbar)
    range_action.setIcon(QtGui.QIcon.fromTheme("measuring_cup"))

    def measure_total() -> None:
        if not data:
            return  # no data

        start_time = min(numpy.min(d[0]).item() for d in data)
        end_time = max(numpy.max(d[0]).item() for d in data)

        display_mean_and_std_dev(start_time, end_time, data, total_action_text,
                                 toolbar)

    def measure_range() -> None:
        QtWidgets.QMessageBox.information(
            toolbar, "Select Measurement Range",
            "Select measurement range.\nUndo = middle click / backspace")

        # grab the points from the user
        points = fig.ginput(n=2, timeout=0, mouse_add=MouseButton.LEFT,
                            mouse_stop=None,  # type: ignore
                            mouse_pop=MouseButton.MIDDLE)
        if len(points) != 2:
            return  # user cancelled
        start_time = min(p[0] for p in points)
        end_time = max(p[0] for p in points)

        # show a box with the freq range
        for ax in axes:
            rect = rects[ax]
            if rect is not None:
                rect.remove()
                rects[ax] = None
                rect = None
            rect = ax.axvspan(start_time, end_time, facecolor='g', alpha=0.5)
            rects[ax] = rect
        fig.canvas.draw()

        display_mean_and_std_dev(start_time, end_time, data, range_action_text,
                                 toolbar)

    total_action.triggered.connect(measure_total)
    range_action.triggered.connect(measure_range)

    toolbar.addAction(total_action)
    toolbar.addAction(range_action)


def time_analysis(parent: QtWidgets.QWidget) -> None:
    ret = util.load_batch(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_index = util.choose_channel(header, parent)
    if channel_index is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, [channel_index], parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent)
    if not ret:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(
        batch_info, parent, chunk=False)

    time, data, _start, _end = sequence_info[channel_index][0]

    show_legend = False

    if data.shape[1] != 1:
        subchannel_index = util.choose_subchannel(
            batch_info[channel_index], allow_all=True, parent=parent)
        if subchannel_index is None:
            return  # user cancelled
        elif subchannel_index != -1:
            data = data[:, subchannel_index:subchannel_index+1]
        else:
            show_legend = True

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    ax.plot(time, data)

    channel_data = batch_info[channel_index]
    subchannel_names = channel_data.channel_decoder.subchannel_names

    mean_std_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                              str, util.UnitInfo]] = []
    for i, name in enumerate(subchannel_names):
        mean_std_data.append((time, data[:, i], name,
                              unit_info[channel_index]))

    ax.set_xlabel("Time (s)")
    unit_str = unit_info[channel_index].utf8
    if unit_str:
        ax.set_ylabel("Magnitude ({})".format(unit_str))
    else:
        ax.set_ylabel("Magnitude")

    ax.get_xaxis().get_major_formatter().set_useOffset(False)  # type: ignore
    ax.get_yaxis().get_major_formatter().set_useOffset(False)  # type: ignore

    if show_legend:
        ax.legend(subchannel_names)

    add_mean_std_dev_summary_buttons(fig, [ax], mean_std_data)

    fig.show()


def synchronous_time_analysis(parent: QtWidgets.QWidget) -> None:
    ret = util.load_batch(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_indexes = util.choose_synchronous_channels(header, parent)
    if channel_indexes is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, channel_indexes, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent, once=True)
    if not ret:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(
        batch_info, parent, chunk=False, unscaled_units=True)

    # group channels by unit type
    plot_groups: list[list[int]] = []
    groups_by_unit_type: dict[int, list[int]] = {}
    for channel_index in sorted(channel_indexes):
        channel = header['channels'][channel_index]
        if channel.unit_type == 0:
            # don't share plots between channels with unknown units
            plot_groups.append([channel_index])
        elif channel.unit_type in groups_by_unit_type:
            # already have a group made
            plot_group = groups_by_unit_type[channel.unit_type]
            plot_group.append(channel_index)
        else:
            # need a new group
            plot_group = [channel_index]
            plot_groups.append(plot_group)
            groups_by_unit_type[channel.unit_type] = plot_group

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax_list = export_script.subplots_wrapper(len(plot_groups),
                                                  squeeze=False, sharex=True)

    mean_std_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                              str, util.UnitInfo]] = []

    for i, ax in enumerate(ax_list[:, 0]):
        plot_group = plot_groups[i]

        legend_strs = []

        for channel_index in plot_group:
            time, data, _start, _end = sequence_info[channel_index][0]
            ax.plot(time, data)

            channel_data = batch_info[channel_index]
            subchannel_names = channel_data.channel_decoder.subchannel_names
            legend_strs.extend(subchannel_names)

            for i, name in enumerate(subchannel_names):
                mean_std_data.append((time, data[:, i], name,
                                      unit_info[channel_index]))

        ax.set_xlabel("Time (s)")
        unit_str = unit_info[plot_group[0]].utf8
        if unit_str:
            ax.set_ylabel("Magnitude ({})".format(unit_str))
        else:
            ax.set_ylabel("Magnitude")

        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)

        ax.legend(legend_strs)

    add_mean_std_dev_summary_buttons(fig, list(ax_list[:, 0]), mean_std_data)

    fig.show()


def overlaid_time_analysis(parent: QtWidgets.QWidget) -> None:
    sequences = []
    subchannel_indexes = []
    unit_types = []
    unit_names = []
    unit_infos: list[util.UnitInfo] = []
    names = []

    while True:
        # load a batch of files
        ret = util.load_batch(parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # choose a channel
        channel_index = util.choose_channel(header, parent)
        if channel_index is None:
            return  # error message displayed, or user cancelled

        batch_info = util.decode_batch(
            file_infos, header, [channel_index], parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        subchannel_index = util.choose_subchannel(
            batch_info[channel_index], allow_all=False, parent=parent)
        if subchannel_index is None:
            return  # user cancelled
        subchannel_indexes.append(subchannel_index)

        batch_info = util.get_datetime_subset(batch_info, parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret = util.warn_about_lost_packets(batch_info, parent)
        if not ret:
            return  # user cancelled

        sequence_info, unit_info = util.sequence_data(
            batch_info, parent, chunk=False, unscaled_units=True)

        sequence = sequence_info[channel_index][0]

        sequences.append(sequence)
        unit_types.append(header['channels'][channel_index].unit_type)
        unit_names.append(unit_info[channel_index].utf8)
        unit_infos.append(unit_info[channel_index])

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.StandardButton.Yes |
                     QtWidgets.QMessageBox.StandardButton.No),
            defaultButton=QtWidgets.QMessageBox.StandardButton.No)
        if ret != QtWidgets.QMessageBox.StandardButton.Yes:
            break

    # group channels by unit type
    plot_groups: list[list[int]] = []
    groups_by_unit_type: dict[int, list[int]] = {}
    for batch_index, unit_type in enumerate(unit_types):
        if unit_type == 0:
            # don't share plots between channels with unknown units
            plot_groups.append([batch_index])
        elif unit_type in groups_by_unit_type:
            # already have a group made
            plot_group = groups_by_unit_type[unit_type]
            plot_group.append(batch_index)
        else:
            # need a new group
            plot_group = [batch_index]
            plot_groups.append(plot_group)
            groups_by_unit_type[unit_type] = plot_group

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax_list = export_script.subplots_wrapper(len(plot_groups),
                                                  squeeze=False, sharex=True)

    mean_std_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                              str, util.UnitInfo]] = []

    for i, ax in enumerate(ax_list[:, 0]):
        plot_group = plot_groups[i]

        for batch_index in plot_group:
            time, data, _start, _end = sequences[batch_index]
            subchannel_index = subchannel_indexes[batch_index]
            ax.plot(time, data[:, subchannel_index], label=names[batch_index])

            mean_std_data.append((time, data[:, subchannel_index],
                                  names[batch_index], unit_infos[batch_index]))

        ax.legend()

        ax.set_xlabel("Time (s)")
        unit_str = unit_names[plot_group[0]]
        if unit_str:
            ax.set_ylabel("Magnitude ({})".format(unit_str))
        else:
            ax.set_ylabel("Magnitude")

        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)

    add_mean_std_dev_summary_buttons(fig, list(ax_list[:, 0]), mean_std_data)

    fig.show()



# Copyright (c) 2018 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import builtins
import gc
import logging
import os.path
from types import ModuleType
from typing import Any, Literal, Optional, overload

import matplotlib.figure
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy
from PySide6 import QtCore, QtGui, QtWidgets
import matplotlib

from . import mondo_rc  # needed for the icon # noqa: F401

logger = logging.getLogger(__name__)


ignored_types = [i for i in builtins.__dict__.values() if isinstance(i, type)]
ignored_types.remove(list)
ignored_types.append(type(None))


def is_ignored_type(value):
    if isinstance(value, QtCore.QObject):
        return True
    for t in ignored_types:
        if type(value) == t:  # noqa: E721
            return True
    return False


class CommandList(list):
    def __init__(self, setup_commands=None):
        self.plotdata = {}
        self._index = 0

        if not setup_commands:
            self.append("import matplotlib")
            self.append("import numpy")
            self.append("from PySide6 import QtGui")
            self.append("")
        else:
            self.extend(setup_commands)
            self.append("")
            self.append("plotdata = numpy.load(__file__ + '.npz')")

    def get_string(self, value):
        if isinstance(value, numpy.ndarray):
            i = 'array_{}'.format(self._index)
            self._index += 1

            self.plotdata[i] = value
            return "plotdata['{}']".format(i)
        else:
            return repr(value)

    def output(self, scriptfile):
        datafile = scriptfile + ".npz"

        logger.debug("Writing script to {}".format(scriptfile))

        with open(scriptfile, "wt", encoding="utf_8") as f:
            for v in self:
                # hack to change fig.show() to plt.show()
                s = str(v)
                if s == "fig.show()":
                    s = "plt.show()"
                f.write(s)
                f.write("\n")

        logger.debug("Writing script data to {}".format(datafile))

        numpy.savez_compressed(datafile, allow_pickle=False, **self.plotdata)

        logger.debug("Finished script output")


class AttributeCommand:
    def __init__(self, cmd_list, base, attr):
        self.cmd_list = cmd_list
        self.base = base
        self.attr = attr

    def __repr__(self):
        return "{}.{}".format(self.base, self.attr)


class CallCommand:
    def __init__(self, cmd_list, base, func, *args, **kwargs):
        self.cmd_list = cmd_list
        self.base = base
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        a = []
        a.extend(self.cmd_list.get_string(x) for x in self.args)
        for k, v in self.kwargs.items():
            a.append(k + "=" + self.cmd_list.get_string(v))
        param_str = ", ".join(a)
        return "{}.{}({})".format(self.base, self.func, param_str)


class GetItemCommand:
    def __init__(self, cmd_list, base, key):
        self.cmd_list = cmd_list
        self.base = base
        self.key = key

    def __repr__(self):
        return "{}[{}]".format(self.base, self.cmd_list.get_string(self.key))


class SetItemCommand:
    def __init__(self, cmd_list, base, key, value):
        self.cmd_list = cmd_list
        self.base = base
        self.key = key
        self.value = value

    def __repr__(self):
        return "{}[{}] = {}".format(self.base,
                                    self.cmd_list.get_string(self.key),
                                    self.cmd_list.get_string(self.value))


class Wrapper:
    def __init__(self, wrapped, name, cmd_list: CommandList):
        self._wrapped = wrapped
        self._name = name
        self._cmd_list = cmd_list

    def __getattribute__(self, attr):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        orig_attr = wrapped.__getattribute__(attr)

        if callable(orig_attr):
            def wrapped_func(*args, **kwargs):
                cmd = CallCommand(cmd_list, name, attr, *args, **kwargs)
                result = orig_attr(*args, **kwargs)
                if is_ignored_type(result):
                    cmd_list.append(cmd)
                    return result
                else:
                    if len(args) + len(kwargs) > 0:
                        cmd_list.append(cmd)
                    return Wrapper(result, cmd, cmd_list)
            return wrapped_func
        else:
            if is_ignored_type(orig_attr):
                return orig_attr
            else:
                new_name = AttributeCommand(cmd_list, name, attr)
                return Wrapper(orig_attr, new_name, cmd_list)

    def __getitem__(self, key):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        new_name = GetItemCommand(cmd_list, name, key)
        value = wrapped.__getitem__(key)
        if is_ignored_type(value):
            return value
        else:
            return Wrapper(value, new_name, cmd_list)

    def __setitem__(self, key, value):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        cmd_list.append(SetItemCommand(cmd_list, name, key, value))
        return wrapped.__setitem__(key, value)

    def __len__(self):
        wrapped = object.__getattribute__(self, '_wrapped')
        return wrapped.__len__()

    def __iter__(self):
        wrapped = object.__getattribute__(self, '_wrapped')
        return wrapped.__iter__()


def get_script_file(
        default_name: str,
        parent: Optional[QtWidgets.QWidget] = None) -> Optional[str]:
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileSaveDirectory")
    if directory and isinstance(directory, str):
        if not os.path.isdir(directory):
            directory = None
    else:
        directory = None

    if not directory:
        directory = ""

    file_and_dir = os.path.join(directory, default_name)

    # ask the user for the file name
    caption = "Save File"
    file_filter = "Python Script (*.py);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(parent, caption, file_and_dir,
                                                file_filter)
    output_path = val[0]

    if output_path:
        # save the directory
        output_dir = os.path.dirname(output_path)
        settings.setValue("fileSaveDirectory", output_dir)
        return output_path
    else:
        return None


def add_export_script_action(figure: Figure, cmd_list: CommandList) -> None:
    toolbar: QtWidgets.QToolBar = figure.canvas.toolbar  # type: ignore
    actionText = QtWidgets.QApplication.translate("ExportScript",
                                                  "Export Script")
    action = QtGui.QAction(actionText, toolbar)
    action.setIcon(QtGui.QIcon.fromTheme("scroll"))

    def handle_export():
        base_file = get_script_file("export", parent=toolbar)
        if not base_file:
            return

        cmd_list.output(base_file)

    action.triggered.connect(handle_export)

    toolbar.addAction(action)


@overload
def subplots_wrapper(
    nrows: int = ...,
    ncols: int = ...,
    *,
    squeeze: Literal[False],
    sharex: bool | Literal["none", "all", "row", "col"] = ...,
    sharey: bool | Literal["none", "all", "row", "col"] = ...,
    subplot_kw: dict = ...,
    gridspec_kw: dict = ...,
    **fig_kw
) -> tuple[Figure, numpy.ndarray]: ...


@overload
def subplots_wrapper(
    nrows: Literal[1] = ...,
    ncols: Literal[1] = ...,
    *,
    sharex: bool | Literal["none", "all", "row", "col"] = ...,
    sharey: bool | Literal["none", "all", "row", "col"] = ...,
    squeeze: Literal[True] = ...,
    subplot_kw: dict = ...,
    gridspec_kw: dict = ...,
    **fig_kw
) -> tuple[Figure, Axes]: ...


def subplots_wrapper(*args, **kwargs) -> tuple[Figure, Any]:
    cmd_list = CommandList(setup_commands)

    cmd_list.append('')

    cmd_list.append('import matplotlib.pyplot as plt')
    import matplotlib.pyplot as plt

    cmd_list.append('')

    # fig, ax = plt.subplots(nrows, ncols, sharex=sharex, sharey=sharey,
    #                        squeeze=squeeze, **fig_kw)
    fig, ax = plt.subplots(*args, **kwargs)

    def handle_close(evt):
        fig.clf()
        cmd_list.clear()
        gc.collect()

    fig.canvas.mpl_connect("close_event", handle_close)

    # set the icon (inside the application only; not in the script)
    icon = QtGui.QIcon()
    icon_reader = QtGui.QImageReader(":/mondo.ico")
    while True:
        pixmap = QtGui.QPixmap.fromImage(icon_reader.read())
        icon.addPixmap(pixmap)
        if not icon_reader.jumpToNextImage():
            break
    fig.canvas.manager.window.setWindowIcon(icon)  # type: ignore

    fig_wrapper = Wrapper(fig, "fig", cmd_list)
    add_export_script_action(fig, cmd_list)

    if isinstance(ax, numpy.ndarray):
        ax_array = ax
        cmd_list.append(CallCommand(
            cmd_list, "fig, ax_array = plt", "subplots", *args, **kwargs))
        ax_wrapper = numpy.empty(ax.shape, dtype=object)
        for i, ax in numpy.ndenumerate(ax_array):
            ax_name = "ax_array[{}]".format(i)
            ax_wrapper[i] = Wrapper(ax, ax_name, cmd_list)
    else:
        cmd_list.append(CallCommand(
            cmd_list, "fig, ax = plt", "subplots", *args, **kwargs))
        ax_wrapper = Wrapper(ax, "ax", cmd_list)

    return fig_wrapper, ax_wrapper  # type: ignore


setup_commands = CommandList()
matplotlib_wrapper: ModuleType = \
    Wrapper(matplotlib, 'matplotlib', setup_commands)  # type: ignore



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preferences.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QGridLayout, QLabel, QRadioButton, QSizePolicy,
    QSpacerItem, QTabWidget, QVBoxLayout, QWidget)

from hyperborea.unit_preferences import UnitPreferencesWidget

class Ui_PreferencesDialog(object):
    def setupUi(self, PreferencesDialog):
        if not PreferencesDialog.objectName():
            PreferencesDialog.setObjectName(u"PreferencesDialog")
        PreferencesDialog.resize(414, 379)
        self.verticalLayout = QVBoxLayout(PreferencesDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.tabWidget = QTabWidget(PreferencesDialog)
        self.tabWidget.setObjectName(u"tabWidget")
        self.interfaceTab = QWidget()
        self.interfaceTab.setObjectName(u"interfaceTab")
        self.gridLayout_2 = QGridLayout(self.interfaceTab)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.ledsLabel_2 = QLabel(self.interfaceTab)
        self.ledsLabel_2.setObjectName(u"ledsLabel_2")
        font = QFont()
        font.setBold(True)
        self.ledsLabel_2.setFont(font)

        self.gridLayout_2.addWidget(self.ledsLabel_2, 0, 0, 1, 2)

        self.horizontalSpacer_6 = QSpacerItem(20, 13, QSizePolicy.Fixed, QSizePolicy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_6, 1, 0, 1, 1)

        self.darkMode = QRadioButton(self.interfaceTab)
        self.darkMode.setObjectName(u"darkMode")

        self.gridLayout_2.addWidget(self.darkMode, 1, 1, 1, 1)

        self.lightMode = QRadioButton(self.interfaceTab)
        self.lightMode.setObjectName(u"lightMode")

        self.gridLayout_2.addWidget(self.lightMode, 2, 1, 1, 1)

        self.verticalSpacer_12 = QSpacerItem(369, 220, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer_12, 3, 0, 1, 2)

        self.gridLayout_2.setColumnStretch(1, 1)
        self.tabWidget.addTab(self.interfaceTab, "")
        self.unitTab = QWidget()
        self.unitTab.setObjectName(u"unitTab")
        self.gridLayout = QGridLayout(self.unitTab)
        self.gridLayout.setObjectName(u"gridLayout")
        self.unitsLabel = QLabel(self.unitTab)
        self.unitsLabel.setObjectName(u"unitsLabel")
        self.unitsLabel.setFont(font)

        self.gridLayout.addWidget(self.unitsLabel, 0, 0, 1, 2)

        self.horizontalSpacer = QSpacerItem(20, 13, QSizePolicy.Fixed, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 1, 0, 1, 1)

        self.verticalSpacer_7 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_7, 4, 0, 1, 2)

        self.unitPreferences = UnitPreferencesWidget(self.unitTab)
        self.unitPreferences.setObjectName(u"unitPreferences")

        self.gridLayout.addWidget(self.unitPreferences, 1, 1, 1, 1)

        self.tabWidget.addTab(self.unitTab, "")

        self.verticalLayout.addWidget(self.tabWidget)

        self.buttonBox = QDialogButtonBox(PreferencesDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        QWidget.setTabOrder(self.tabWidget, self.buttonBox)

        self.retranslateUi(PreferencesDialog)
        self.buttonBox.accepted.connect(PreferencesDialog.accept)
        self.buttonBox.rejected.connect(PreferencesDialog.reject)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(PreferencesDialog)
    # setupUi

    def retranslateUi(self, PreferencesDialog):
        PreferencesDialog.setWindowTitle(QCoreApplication.translate("PreferencesDialog", u"Preferences", None))
        self.ledsLabel_2.setText(QCoreApplication.translate("PreferencesDialog", u"Display Mode", None))
        self.darkMode.setText(QCoreApplication.translate("PreferencesDialog", u"Dark Mode", None))
        self.lightMode.setText(QCoreApplication.translate("PreferencesDialog", u"Light Mode", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.interfaceTab), QCoreApplication.translate("PreferencesDialog", u"Interface", None))
        self.unitsLabel.setText(QCoreApplication.translate("PreferencesDialog", u"Display Units", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.unitTab), QCoreApplication.translate("PreferencesDialog", u"Units", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'psd_options.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDoubleSpinBox, QFormLayout,
    QHBoxLayout, QLabel, QSizePolicy, QSpacerItem,
    QSpinBox, QWidget)

class Ui_PSDOptionsWidget(object):
    def setupUi(self, PSDOptionsWidget):
        if not PSDOptionsWidget.objectName():
            PSDOptionsWidget.setObjectName(u"PSDOptionsWidget")
        PSDOptionsWidget.resize(208, 191)
        self.formLayout = QFormLayout(PSDOptionsWidget)
        self.formLayout.setContentsMargins(0, 0, 0, 0)
        self.formLayout.setObjectName(u"formLayout")
        self.fftPointsLabel = QLabel(PSDOptionsWidget)
        self.fftPointsLabel.setObjectName(u"fftPointsLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.fftPointsLabel)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_4)

        self.fftPoints = QComboBox(PSDOptionsWidget)
        self.fftPoints.setObjectName(u"fftPoints")

        self.horizontalLayout_4.addWidget(self.fftPoints)


        self.formLayout.setLayout(0, QFormLayout.FieldRole, self.horizontalLayout_4)

        self.durationLabel = QLabel(PSDOptionsWidget)
        self.durationLabel.setObjectName(u"durationLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.durationLabel)

        self.duration = QLabel(PSDOptionsWidget)
        self.duration.setObjectName(u"duration")
        self.duration.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.duration)

        self.resolutionLabel = QLabel(PSDOptionsWidget)
        self.resolutionLabel.setObjectName(u"resolutionLabel")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.resolutionLabel)

        self.resolution = QLabel(PSDOptionsWidget)
        self.resolution.setObjectName(u"resolution")
        self.resolution.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.resolution)

        self.overlapPercentLabel = QLabel(PSDOptionsWidget)
        self.overlapPercentLabel.setObjectName(u"overlapPercentLabel")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.overlapPercentLabel)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)

        self.overlapPercent = QDoubleSpinBox(PSDOptionsWidget)
        self.overlapPercent.setObjectName(u"overlapPercent")
        self.overlapPercent.setMinimumSize(QSize(75, 0))
        self.overlapPercent.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.overlapPercent.setDecimals(0)
        self.overlapPercent.setMaximum(95.000000000000000)
        self.overlapPercent.setSingleStep(5.000000000000000)
        self.overlapPercent.setValue(75.000000000000000)

        self.horizontalLayout_3.addWidget(self.overlapPercent)


        self.formLayout.setLayout(3, QFormLayout.FieldRole, self.horizontalLayout_3)

        self.overlapPointsLabel = QLabel(PSDOptionsWidget)
        self.overlapPointsLabel.setObjectName(u"overlapPointsLabel")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.overlapPointsLabel)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_5)

        self.overlapPoints = QSpinBox(PSDOptionsWidget)
        self.overlapPoints.setObjectName(u"overlapPoints")
        self.overlapPoints.setMinimumSize(QSize(75, 0))
        self.overlapPoints.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.horizontalLayout_5.addWidget(self.overlapPoints)


        self.formLayout.setLayout(4, QFormLayout.FieldRole, self.horizontalLayout_5)

        self.windowCountLabel = QLabel(PSDOptionsWidget)
        self.windowCountLabel.setObjectName(u"windowCountLabel")

        self.formLayout.setWidget(5, QFormLayout.LabelRole, self.windowCountLabel)

        self.windowCount = QLabel(PSDOptionsWidget)
        self.windowCount.setObjectName(u"windowCount")
        self.windowCount.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.formLayout.setWidget(5, QFormLayout.FieldRole, self.windowCount)

        self.windowFunctionLabel = QLabel(PSDOptionsWidget)
        self.windowFunctionLabel.setObjectName(u"windowFunctionLabel")

        self.formLayout.setWidget(6, QFormLayout.LabelRole, self.windowFunctionLabel)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.windowFunction = QComboBox(PSDOptionsWidget)
        self.windowFunction.setObjectName(u"windowFunction")

        self.horizontalLayout_2.addWidget(self.windowFunction)


        self.formLayout.setLayout(6, QFormLayout.FieldRole, self.horizontalLayout_2)

        self.detrendMethodLabel = QLabel(PSDOptionsWidget)
        self.detrendMethodLabel.setObjectName(u"detrendMethodLabel")

        self.formLayout.setWidget(7, QFormLayout.LabelRole, self.detrendMethodLabel)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.detrendMethod = QComboBox(PSDOptionsWidget)
        self.detrendMethod.setObjectName(u"detrendMethod")

        self.horizontalLayout.addWidget(self.detrendMethod)


        self.formLayout.setLayout(7, QFormLayout.FieldRole, self.horizontalLayout)


        self.retranslateUi(PSDOptionsWidget)

        QMetaObject.connectSlotsByName(PSDOptionsWidget)
    # setupUi

    def retranslateUi(self, PSDOptionsWidget):
        PSDOptionsWidget.setWindowTitle(QCoreApplication.translate("PSDOptionsWidget", u"PSD Options", None))
        self.fftPointsLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"FFT Window Size", None))
        self.durationLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Window Duration", None))
        self.duration.setText(QCoreApplication.translate("PSDOptionsWidget", u"600 s", None))
        self.resolutionLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Frequency Resolution", None))
        self.resolution.setText(QCoreApplication.translate("PSDOptionsWidget", u"0.1 Hz", None))
        self.overlapPercentLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Window Overlap Percent", None))
        self.overlapPercent.setSuffix(QCoreApplication.translate("PSDOptionsWidget", u" %", None))
        self.overlapPointsLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Window Overlap Points", None))
        self.windowCountLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Number of Windows", None))
        self.windowCount.setText(QCoreApplication.translate("PSDOptionsWidget", u"1", None))
        self.windowFunctionLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Window Function", None))
        self.detrendMethodLabel.setText(QCoreApplication.translate("PSDOptionsWidget", u"Detrend Method", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'about.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFormLayout, QLabel, QSizePolicy, QVBoxLayout,
    QWidget)

class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        if not AboutDialog.objectName():
            AboutDialog.setObjectName(u"AboutDialog")
        AboutDialog.resize(505, 506)
        self.verticalLayout = QVBoxLayout(AboutDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setVerticalSpacing(20)
        self.label = QLabel(AboutDialog)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)

        self.softwareLabel = QLabel(AboutDialog)
        self.softwareLabel.setObjectName(u"softwareLabel")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.softwareLabel.sizePolicy().hasHeightForWidth())
        self.softwareLabel.setSizePolicy(sizePolicy)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.softwareLabel)

        self.label_4 = QLabel(AboutDialog)
        self.label_4.setObjectName(u"label_4")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_4)

        self.label_9 = QLabel(AboutDialog)
        self.label_9.setObjectName(u"label_9")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_9.sizePolicy().hasHeightForWidth())
        self.label_9.setSizePolicy(sizePolicy1)
        self.label_9.setWordWrap(True)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.label_9)

        self.label_5 = QLabel(AboutDialog)
        self.label_5.setObjectName(u"label_5")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_5)

        self.label_10 = QLabel(AboutDialog)
        self.label_10.setObjectName(u"label_10")
        sizePolicy.setHeightForWidth(self.label_10.sizePolicy().hasHeightForWidth())
        self.label_10.setSizePolicy(sizePolicy)
        self.label_10.setWordWrap(True)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.label_10)

        self.label_6 = QLabel(AboutDialog)
        self.label_6.setObjectName(u"label_6")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.label_6)

        self.label_11 = QLabel(AboutDialog)
        self.label_11.setObjectName(u"label_11")
        sizePolicy.setHeightForWidth(self.label_11.sizePolicy().hasHeightForWidth())
        self.label_11.setSizePolicy(sizePolicy)
        self.label_11.setWordWrap(True)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.label_11)

        self.label_7 = QLabel(AboutDialog)
        self.label_7.setObjectName(u"label_7")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.label_7)

        self.label_12 = QLabel(AboutDialog)
        self.label_12.setObjectName(u"label_12")
        sizePolicy.setHeightForWidth(self.label_12.sizePolicy().hasHeightForWidth())
        self.label_12.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.label_12.setFont(font)
        self.label_12.setWordWrap(True)

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.label_12)


        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(AboutDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(AboutDialog)
        self.buttonBox.accepted.connect(AboutDialog.accept)
        self.buttonBox.rejected.connect(AboutDialog.reject)

        QMetaObject.connectSlotsByName(AboutDialog)
    # setupUi

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(QCoreApplication.translate("AboutDialog", u"About", None))
        self.label.setText(QCoreApplication.translate("AboutDialog", u"Software:", None))
        self.softwareLabel.setText(QCoreApplication.translate("AboutDialog", u"Mondo\n"
"Version: {}\n"
"Build Date: {}", None))
        self.label_4.setText(QCoreApplication.translate("AboutDialog", u"Support:", None))
        self.label_9.setText(QCoreApplication.translate("AboutDialog", u"This software is provided by EPRI \"AS IS\" and without customer support beyond such embodiments within the distribution of this software that may or may not provide such support.", None))
        self.label_5.setText(QCoreApplication.translate("AboutDialog", u"Copyright:", None))
        self.label_10.setText(QCoreApplication.translate("AboutDialog", u"Copyright (c) 2016 Electric Power Research Institute, Inc.\n"
"Copyright (c) 2021 Suprock Technologies, LLC\n"
"All rights reserved.\n"
"\n"
"Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.", None))
        self.label_6.setText(QCoreApplication.translate("AboutDialog", u"Developed by:", None))
        self.label_11.setText(QCoreApplication.translate("AboutDialog", u"Suprock Technologies\n"
"45 Scott Hill Rd\n"
"Warren, NH 03279", None))
        self.label_7.setText(QCoreApplication.translate("AboutDialog", u"Disclaimer:", None))
        self.label_12.setText(QCoreApplication.translate("AboutDialog", u"THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EPRI BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.", None))
    # retranslateUi




# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import importlib
import logging
import os
import subprocess
import sys
import tempfile
import threading
from typing import Callable, Concatenate, Optional, ParamSpec

from PySide6 import QtCore, QtGui, QtWidgets

import asphodel
from hyperborea.dark_mode import set_style
import hyperborea.download
from hyperborea.preferences import read_bool_setting

from .ui.ui_main import Ui_MondoMainWindow
from .about import AboutDialog
from .analysis import csv, file, psd, spectrogram, time
from .preferences import PreferencesDialog

logger = logging.getLogger(__name__)


P = ParamSpec('P')


class MondoMainWindow(Ui_MondoMainWindow, QtWidgets.QMainWindow):
    find_update_finished = QtCore.Signal(object)
    update_download_finished = QtCore.Signal(object)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.update_progress_lock = threading.Lock()

        self.settings = QtCore.QSettings()
        self.update_style()

        self.setupUi(self)

        self.extra_ui_setup()
        self.setup_callbacks()
        self.setup_update_actions()

    def extra_ui_setup(self) -> None:
        app_name = QtWidgets.QApplication.applicationName()
        is_frozen = getattr(sys, 'frozen', False)
        if is_frozen:
            version = QtWidgets.QApplication.applicationVersion()
            title = self.tr("{} ({})").format(app_name, version)
        else:
            title = self.tr("{} (dev)").format(app_name)
        self.setWindowTitle(title)

        self.update_progress = QtWidgets.QProgressDialog("", "", 0, 100)
        self.update_progress.setLabelText(self.tr(""))
        self.update_progress.setWindowTitle(self.tr("Check for Update"))
        self.update_progress.setCancelButton(None)  # type: ignore
        self.update_progress.setWindowModality(
            QtCore.Qt.WindowModality.WindowModal)
        self.update_progress.setMinimumDuration(0)
        self.update_progress.setAutoReset(False)
        self.update_progress.reset()

    def setup_callbacks(self) -> None:
        self.timeButton.clicked.connect(self.time_analysis)
        self.synchronousTimeButton.clicked.connect(
            self.synchronous_time_analysis)
        self.overlaidTimeButton.clicked.connect(self.overlaid_time_analysis)
        self.spectrogramButton.clicked.connect(self.spectrogram_analysis)
        self.psdButton.clicked.connect(self.psd_analysis)
        self.singleChannelPsdButton.clicked.connect(
            self.single_channel_psd_analysis)
        self.singleSlicePsdButton.clicked.connect(
            self.single_slice_psd_analysis)
        self.overlaidPsdButton.clicked.connect(self.overlaid_psd_analysis)
        self.csvButton.clicked.connect(self.csv_export)
        self.csvDownsampledButton.clicked.connect(self.csv_downsampled_export)
        self.fileInformationButton.clicked.connect(self.file_information)
        self.viewSettingsButton.clicked.connect(self.view_settings)
        self.rawExportButton.clicked.connect(self.raw_export)
        self.splitFileButton.clicked.connect(self.split_file)

        self.actionPreferences.triggered.connect(self.show_preferences)

        self.actionAbout.triggered.connect(self.show_about)
        self.actionAboutLibraries.triggered.connect(self.show_about_libraries)
        self.actionShowLogs.triggered.connect(self.show_log_dir)

        self.actionUpdateLatestStable.triggered.connect(
            self.update_latest_stable)
        self.actionUpdateCurrentBranch.triggered.connect(
            self.update_current_branch)
        self.actionUpdateSpecificBranch.triggered.connect(
            self.update_specific_branch)
        self.actionUpdateSpecificCommit.triggered.connect(
            self.update_specific_commit)

        self.software_finder = hyperborea.download.SoftwareFinder(logger)
        self.software_finder.completed.connect(self.update_finder_completed)
        self.software_finder.error.connect(self.update_finder_error)
        self.ref_finder = hyperborea.download.RefFinder(logger)
        self.ref_finder.completed.connect(self.ref_finder_completed)
        self.ref_finder.error.connect(self.ref_finder_error)
        self.software_downloader = hyperborea.download.Downloader(logger)
        self.software_downloader.update.connect(self.update_progress_cb)
        self.software_downloader.completed.connect(
            self.software_download_completed)
        self.software_downloader.error.connect(self.software_download_error)

    def setup_update_actions(self) -> None:
        is_frozen = getattr(sys, 'frozen', False)

        valid_info = False
        if is_frozen:
            # load the build_info.txt
            main_dir = os.path.dirname(sys.executable)
            build_info_filename = os.path.join(main_dir, "build_info.txt")
            try:
                with open(build_info_filename, "r", encoding="utf-8") as f:
                    lines = f.readlines()
                    self.branch_name = lines[0].strip()
                    self.commit_hash = lines[1].strip()
                    self.build_key = lines[2].strip()
                    valid_info = True
            except Exception:
                logger.exception('Could not read build_info.txt')

        if not valid_info:
            self.menuCheckForUpdates.setEnabled(False)
            self.menuCheckForUpdates.setTitle(self.tr("Not Updatable"))
            self.actionUpdateLatestStable.setEnabled(False)
            self.actionUpdateCurrentBranch.setEnabled(False)
            self.actionUpdateSpecificBranch.setEnabled(False)
            self.actionUpdateSpecificCommit.setEnabled(False)
        else:
            if self.branch_name == "master":
                # master is latest stable
                self.actionUpdateCurrentBranch.setEnabled(False)
                self.actionUpdateCurrentBranch.setVisible(False)
            else:
                action_str = self.tr("Latest {}").format(self.branch_name)
                self.actionUpdateCurrentBranch.setText(action_str)

    def find_update(self, branch=None, commit=None, fallback_branch=None):
        self.update_fallback_branch = fallback_branch

        self.update_progress.setMinimum(0)
        self.update_progress.setMaximum(0)
        self.update_progress.setValue(0)
        self.update_progress.setLabelText(self.tr("Checking for update..."))
        self.update_progress.forceShow()

        self.software_finder.find_software("mondo", self.build_key, branch,
                                           commit)

    @QtCore.Slot(str)
    def update_finder_error(self, error_str: str):
        if self.update_fallback_branch:
            self.find_update(branch=self.update_fallback_branch)
        else:
            self.update_progress.reset()
            QtWidgets.QMessageBox.critical(self, self.tr("Error"), error_str)

    @QtCore.Slot(object)
    def update_finder_completed(self, params) -> None:
        self.update_progress.reset()
        url, commit, ready = params

        if commit is not None and commit == self.commit_hash:
            logger.info("Up to date with commit %s", commit)
            QtWidgets.QMessageBox.information(
                self, self.tr("Up to date"),
                self.tr("Already running this version"))
            return

        if not ready:
            logger.info("Update is not ready")
            QtWidgets.QMessageBox.information(self,
                                              self.tr("Update not ready"),
                                              self.tr("Update is not ready"))
            return

        # ask if the user wants to proceed
        ret = QtWidgets.QMessageBox.question(
            self, self.tr("Update?"), self.tr("Update available. Update now?"),
            QtWidgets.QMessageBox.StandardButton.Yes |
            QtWidgets.QMessageBox.StandardButton.No)
        if ret != QtWidgets.QMessageBox.StandardButton.Yes:
            return

        logger.info("Downloading update from %s", url)

        self.update_progress.setMinimum(0)
        self.update_progress.setMaximum(0)
        self.update_progress.setValue(0)
        self.update_progress.setLabelText(self.tr("Downloading update..."))
        self.update_progress.forceShow()

        fd, filename = tempfile.mkstemp(".exe", "setup-", text=False)
        file = os.fdopen(fd, "wb")
        file.filename = filename  # type: ignore
        self.software_downloader.start_download(url, file)

    @QtCore.Slot(int, int)
    def update_progress_cb(self, written_bytes: int, total_length: int):
        if total_length != 0:
            self.update_progress.setMinimum(0)
            self.update_progress.setMaximum(total_length)
            self.update_progress.setValue(written_bytes)

    @QtCore.Slot(object, str)
    def software_download_error(self, file, error_str: str):
        self.update_progress.reset()
        file.close()
        QtWidgets.QMessageBox.critical(self, self.tr("Error"), error_str)
        os.unlink(file.filename)

    @QtCore.Slot(str, object)
    def software_download_completed(self, _url: str, file) -> None:
        self.update_progress.reset()
        file.close()

        # run the intstaller
        subprocess.Popen([
            file.filename, '/silent', "/DeleteInstaller=Yes", "/SP-",
            "/SUPPRESSMSGBOXES", "/NORESTART", "/NOCANCEL"
        ])

        # close the application (though installer will force kill regardless)
        self.close()

    @QtCore.Slot()
    def update_latest_stable(self):
        self.find_update(branch="master")

    @QtCore.Slot()
    def update_current_branch(self):
        if self.branch_name in ("master", "develop"):
            fallback = None
        else:
            fallback = "develop"

        self.find_update(branch=self.branch_name, fallback_branch=fallback)

    @QtCore.Slot()
    def update_specific_branch(self):
        self.update_progress.setMinimum(0)
        self.update_progress.setMaximum(0)
        self.update_progress.setValue(0)
        self.update_progress.setLabelText(self.tr("Collecting branches..."))
        self.update_progress.forceShow()

        self.ref_finder.get_software_refs("mondo")

    @QtCore.Slot()
    def ref_finder_error(self):
        self.ref_finder_completed([])

    @QtCore.Slot(object)
    def ref_finder_completed(self, refs):
        self.update_progress.reset()

        default_branch = self.branch_name
        branch_choices = [default_branch]
        if default_branch != "master":
            branch_choices.append("master")
        if default_branch != "develop":
            branch_choices.append("develop")

        ref_names = []
        for ref in refs:
            ref_name = ref.get("name", None)
            if ref_name and ref_name not in branch_choices:
                ref_names.append(ref_name)

        if ref_names:
            label = self.tr("Branch:")
        else:
            label = self.tr("No list of branches!\nBranch:")

        branch_choices.extend(sorted(ref_names,
                                     key=hyperborea.download.ref_sort_key))

        branch, ok = QtWidgets.QInputDialog.getItem(self, self.tr("Branch"),
                                                    label, branch_choices, 0,
                                                    True)
        if not ok:
            return

        branch = branch.strip()

        self.find_update(branch=branch)

    @QtCore.Slot()
    def update_specific_commit(self):
        commit, ok = QtWidgets.QInputDialog.getText(
            self, self.tr("Commit"), self.tr("Commit:"),
            QtWidgets.QLineEdit.EchoMode.Normal, "")
        if not ok:
            return

        commit = commit.strip()

        self.find_update(commit=commit)

    @QtCore.Slot()
    def show_about(self) -> None:
        dialog = AboutDialog(self)
        dialog.exec()

    def _get_version(self, library: str) -> str:
        try:
            lib = importlib.import_module(library)
            return lib.__version__
        except (AttributeError, ImportError):
            return "ERROR"

    @QtCore.Slot()
    def show_about_libraries(self) -> None:
        libraries = [
            "hyperborea",
            "matplotlib",
            "numpy",
            "PySide6",
            "requests",
            "scipy",
        ]
        vers: dict[str, str] = {}
        for lib in libraries:
            vers[lib] = self._get_version(lib)

        # special case for asphodel and asphodel_py
        vers["asphodel_py"] = self._get_version("asphodel")
        vers["asphodel"] = asphodel.build_info

        # python version (sys.version is too long)
        is_64bit = sys.maxsize > (2 ** 32)
        bit_str = "64 bit" if is_64bit else "32 bit"
        python_ver = ".".join(map(str, sys.version_info[:3]))
        python_str = "{} ({} {})".format(python_ver, sys.platform, bit_str)
        vers['python'] = python_str

        s = "\n".join(k + ": " + vers[k] for k in sorted(vers, key=str.lower))
        QtWidgets.QMessageBox.information(self, "About Libraries", s)

    @QtCore.Slot()
    def show_log_dir(self) -> None:
        logdir = os.path.abspath(
            QtCore.QStandardPaths.writableLocation(
                QtCore.QStandardPaths.StandardLocation.AppLocalDataLocation))
        logfile = os.path.join(logdir, "main.log")

        if sys.platform == "win32":
            subprocess.Popen(['explorer', '/select,', logfile])
        elif sys.platform == "darwin":
            args = [
                "osascript",
                "-e", 'tell application "Finder"',
                "-e", "activate",
                "-e", f'select POSIX file "{logfile}"',
                "-e", "end tell",
            ]
            subprocess.Popen(args)
        else:
            url = QtCore.QUrl.fromLocalFile(logdir)
            QtGui.QDesktopServices.openUrl(url)

    def do_analysis(
            self, analysis: Callable[Concatenate[QtWidgets.QWidget, P], None],
            *args: P.args, **kwargs: P.kwargs) -> None:
        try:
            analysis(self, *args, **kwargs)
        except Exception:
            logger.exception("Uncaught exception")
            QtWidgets.QMessageBox.critical(
                self, "Error", "Uncaught exception. See log for details.")

    @QtCore.Slot()
    def time_analysis(self) -> None:
        self.do_analysis(time.time_analysis)

    @QtCore.Slot()
    def synchronous_time_analysis(self) -> None:
        self.do_analysis(time.synchronous_time_analysis)

    @QtCore.Slot()
    def overlaid_time_analysis(self) -> None:
        self.do_analysis(time.overlaid_time_analysis)

    @QtCore.Slot()
    def spectrogram_analysis(self) -> None:
        self.do_analysis(spectrogram.spectrogram_analysis)

    @QtCore.Slot()
    def psd_analysis(self) -> None:
        self.do_analysis(psd.psd_analysis)

    @QtCore.Slot()
    def single_channel_psd_analysis(self) -> None:
        self.do_analysis(psd.single_channel_psd_analysis)

    @QtCore.Slot()
    def single_slice_psd_analysis(self) -> None:
        self.do_analysis(psd.single_slice_psd_analysis)

    @QtCore.Slot()
    def overlaid_psd_analysis(self) -> None:
        self.do_analysis(psd.overlaid_psd_analysis)

    @QtCore.Slot()
    def csv_export(self) -> None:
        self.do_analysis(csv.csv_export)

    @QtCore.Slot()
    def csv_downsampled_export(self) -> None:
        self.do_analysis(csv.csv_export, downsample=True)

    @QtCore.Slot()
    def file_information(self) -> None:
        self.do_analysis(file.file_information)

    @QtCore.Slot()
    def view_settings(self) -> None:
        self.do_analysis(file.view_settings)

    @QtCore.Slot()
    def raw_export(self) -> None:
        self.do_analysis(file.raw_export)

    @QtCore.Slot()
    def split_file(self) -> None:
        self.do_analysis(file.split_file)

    @QtCore.Slot()
    def show_preferences(self) -> None:
        dialog = PreferencesDialog(self)
        dialog.exec()
        self.update_style()

    def update_style(self) -> None:
        dark_mode = read_bool_setting(self.settings, "DarkMode", True)
        set_style(QtWidgets.QApplication.instance(), dark_mode)  # type: ignore



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'datetime_subset.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDateTimeEdit, QDialog,
    QDialogButtonBox, QFormLayout, QGridLayout, QLabel,
    QRadioButton, QSizePolicy, QSpacerItem, QSpinBox,
    QWidget)

class Ui_DateTimeSubsetDialog(object):
    def setupUi(self, DateTimeSubsetDialog):
        if not DateTimeSubsetDialog.objectName():
            DateTimeSubsetDialog.setObjectName(u"DateTimeSubsetDialog")
        DateTimeSubsetDialog.resize(283, 284)
        self.gridLayout = QGridLayout(DateTimeSubsetDialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.verticalSpacer = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 4, 0, 1, 3)

        self.useAllData = QRadioButton(DateTimeSubsetDialog)
        self.useAllData.setObjectName(u"useAllData")
        self.useAllData.setChecked(True)

        self.gridLayout.addWidget(self.useAllData, 0, 0, 1, 3)

        self.buttonBox = QDialogButtonBox(DateTimeSubsetDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 3)

        self.formLayout_2 = QFormLayout()
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.formLayout_2.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.startLabel = QLabel(DateTimeSubsetDialog)
        self.startLabel.setObjectName(u"startLabel")
        self.startLabel.setEnabled(False)

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.startLabel)

        self.startDateTime = QDateTimeEdit(DateTimeSubsetDialog)
        self.startDateTime.setObjectName(u"startDateTime")
        self.startDateTime.setEnabled(False)
        self.startDateTime.setWrapping(True)

        self.formLayout_2.setWidget(0, QFormLayout.FieldRole, self.startDateTime)

        self.endLabel = QLabel(DateTimeSubsetDialog)
        self.endLabel.setObjectName(u"endLabel")
        self.endLabel.setEnabled(False)

        self.formLayout_2.setWidget(1, QFormLayout.LabelRole, self.endLabel)

        self.endDateTime = QDateTimeEdit(DateTimeSubsetDialog)
        self.endDateTime.setObjectName(u"endDateTime")
        self.endDateTime.setEnabled(False)
        self.endDateTime.setWrapping(True)

        self.formLayout_2.setWidget(1, QFormLayout.FieldRole, self.endDateTime)

        self.durationLabel = QLabel(DateTimeSubsetDialog)
        self.durationLabel.setObjectName(u"durationLabel")
        self.durationLabel.setEnabled(False)

        self.formLayout_2.setWidget(4, QFormLayout.LabelRole, self.durationLabel)

        self.label = QLabel(DateTimeSubsetDialog)
        self.label.setObjectName(u"label")
        self.label.setEnabled(False)

        self.formLayout_2.setWidget(2, QFormLayout.LabelRole, self.label)

        self.label_2 = QLabel(DateTimeSubsetDialog)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setEnabled(False)

        self.formLayout_2.setWidget(3, QFormLayout.LabelRole, self.label_2)

        self.startSeconds = QSpinBox(DateTimeSubsetDialog)
        self.startSeconds.setObjectName(u"startSeconds")
        self.startSeconds.setEnabled(False)

        self.formLayout_2.setWidget(2, QFormLayout.FieldRole, self.startSeconds)

        self.endSeconds = QSpinBox(DateTimeSubsetDialog)
        self.endSeconds.setObjectName(u"endSeconds")
        self.endSeconds.setEnabled(False)

        self.formLayout_2.setWidget(3, QFormLayout.FieldRole, self.endSeconds)

        self.duration = QSpinBox(DateTimeSubsetDialog)
        self.duration.setObjectName(u"duration")
        self.duration.setEnabled(False)

        self.formLayout_2.setWidget(4, QFormLayout.FieldRole, self.duration)


        self.gridLayout.addLayout(self.formLayout_2, 3, 1, 1, 2)

        self.horizontalSpacer = QSpacerItem(13, 13, QSizePolicy.Fixed, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 3, 0, 1, 1)

        self.useSubset = QRadioButton(DateTimeSubsetDialog)
        self.useSubset.setObjectName(u"useSubset")

        self.gridLayout.addWidget(self.useSubset, 2, 0, 1, 3)

        self.formLayout_4 = QFormLayout()
        self.formLayout_4.setObjectName(u"formLayout_4")
        self.allStartLabel = QLabel(DateTimeSubsetDialog)
        self.allStartLabel.setObjectName(u"allStartLabel")

        self.formLayout_4.setWidget(0, QFormLayout.LabelRole, self.allStartLabel)

        self.allStart = QLabel(DateTimeSubsetDialog)
        self.allStart.setObjectName(u"allStart")

        self.formLayout_4.setWidget(0, QFormLayout.FieldRole, self.allStart)

        self.allEndLabel = QLabel(DateTimeSubsetDialog)
        self.allEndLabel.setObjectName(u"allEndLabel")

        self.formLayout_4.setWidget(1, QFormLayout.LabelRole, self.allEndLabel)

        self.allEnd = QLabel(DateTimeSubsetDialog)
        self.allEnd.setObjectName(u"allEnd")

        self.formLayout_4.setWidget(1, QFormLayout.FieldRole, self.allEnd)

        self.allDurationLabel = QLabel(DateTimeSubsetDialog)
        self.allDurationLabel.setObjectName(u"allDurationLabel")

        self.formLayout_4.setWidget(2, QFormLayout.LabelRole, self.allDurationLabel)

        self.allDuration = QLabel(DateTimeSubsetDialog)
        self.allDuration.setObjectName(u"allDuration")

        self.formLayout_4.setWidget(2, QFormLayout.FieldRole, self.allDuration)


        self.gridLayout.addLayout(self.formLayout_4, 1, 1, 1, 2)

        QWidget.setTabOrder(self.useAllData, self.useSubset)
        QWidget.setTabOrder(self.useSubset, self.startDateTime)
        QWidget.setTabOrder(self.startDateTime, self.endDateTime)
        QWidget.setTabOrder(self.endDateTime, self.buttonBox)

        self.retranslateUi(DateTimeSubsetDialog)
        self.buttonBox.accepted.connect(DateTimeSubsetDialog.accept)
        self.buttonBox.rejected.connect(DateTimeSubsetDialog.reject)
        self.useSubset.toggled.connect(self.startLabel.setEnabled)
        self.useSubset.toggled.connect(self.startDateTime.setEnabled)
        self.useSubset.toggled.connect(self.endLabel.setEnabled)
        self.useSubset.toggled.connect(self.endDateTime.setEnabled)
        self.useSubset.toggled.connect(self.durationLabel.setEnabled)
        self.useAllData.toggled.connect(self.allStartLabel.setEnabled)
        self.useAllData.toggled.connect(self.allStart.setEnabled)
        self.useAllData.toggled.connect(self.allEndLabel.setEnabled)
        self.useAllData.toggled.connect(self.allEnd.setEnabled)
        self.useAllData.toggled.connect(self.allDurationLabel.setEnabled)
        self.useAllData.toggled.connect(self.allDuration.setEnabled)
        self.useSubset.toggled.connect(self.label.setEnabled)
        self.useSubset.toggled.connect(self.label_2.setEnabled)
        self.useSubset.toggled.connect(self.startSeconds.setEnabled)
        self.useSubset.toggled.connect(self.endSeconds.setEnabled)
        self.useSubset.toggled.connect(self.duration.setEnabled)

        QMetaObject.connectSlotsByName(DateTimeSubsetDialog)
    # setupUi

    def retranslateUi(self, DateTimeSubsetDialog):
        DateTimeSubsetDialog.setWindowTitle(QCoreApplication.translate("DateTimeSubsetDialog", u"Select Data Range", None))
        self.useAllData.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Use all data", None))
        self.startLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Start Time (UTC):", None))
        self.startDateTime.setDisplayFormat(QCoreApplication.translate("DateTimeSubsetDialog", u"yyyy-MM-ddThh:mm:ssZ", None))
        self.endLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"End Time (UTC):", None))
        self.endDateTime.setDisplayFormat(QCoreApplication.translate("DateTimeSubsetDialog", u"yyyy-MM-ddThh:mm:ssZ", None))
        self.durationLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Duration:", None))
        self.label.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Relative Start Time", None))
        self.label_2.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Relative End Time", None))
        self.startSeconds.setSuffix(QCoreApplication.translate("DateTimeSubsetDialog", u" s", None))
        self.endSeconds.setSuffix(QCoreApplication.translate("DateTimeSubsetDialog", u" s", None))
        self.duration.setSuffix(QCoreApplication.translate("DateTimeSubsetDialog", u" s", None))
        self.useSubset.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Use data within range", None))
        self.allStartLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Start Time (UTC):", None))
        self.allStart.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"1/1/2000 12:00:00 AM", None))
        self.allEndLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"End Time (UTC):", None))
        self.allEnd.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"1/1/2000 12:00:00 AM", None))
        self.allDurationLabel.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"Duration:", None))
        self.allDuration.setText(QCoreApplication.translate("DateTimeSubsetDialog", u"0 s", None))
    # retranslateUi




import logging
from typing import Optional

from PySide6 import QtCore, QtWidgets

from hyperborea.dark_mode import set_style
from hyperborea.preferences import (read_bool_setting, write_bool_setting)

from .ui.ui_preferences import Ui_PreferencesDialog

logger = logging.getLogger(__name__)


original_palette = None


class PreferencesDialog(Ui_PreferencesDialog, QtWidgets.QDialog):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.settings = QtCore.QSettings()

        self.setupUi(self)

        # this is easily forgotten in Qt Designer
        self.tabWidget.setCurrentIndex(0)

        self.accepted.connect(self.write_settings)
        self.darkMode.toggled.connect(self.dark_mode_updated)
        self.lightMode.toggled.connect(self.dark_mode_updated)

        self.read_settings()

    @QtCore.Slot()
    def dark_mode_updated(self) -> None:
        dark_mode = self.darkMode.isChecked()
        set_style(QtWidgets.QApplication.instance(), dark_mode)  # type: ignore

    def read_settings(self):
        self.unitPreferences.read_settings()

        dark_mode = read_bool_setting(self.settings, "DarkMode", True)
        if dark_mode:
            self.darkMode.setChecked(True)
        else:
            self.lightMode.setChecked(True)

    @QtCore.Slot()
    def write_settings(self) -> None:
        self.unitPreferences.write_settings()

        dark_mode = self.darkMode.isChecked()
        write_bool_setting(self.settings, "DarkMode", dark_mode)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QGroupBox, QMainWindow,
    QMenu, QMenuBar, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_MondoMainWindow(object):
    def setupUi(self, MondoMainWindow):
        if not MondoMainWindow.objectName():
            MondoMainWindow.setObjectName(u"MondoMainWindow")
        MondoMainWindow.resize(617, 366)
        self.actionAbout = QAction(MondoMainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.actionExit = QAction(MondoMainWindow)
        self.actionExit.setObjectName(u"actionExit")
        self.actionAboutLibraries = QAction(MondoMainWindow)
        self.actionAboutLibraries.setObjectName(u"actionAboutLibraries")
        self.actionUpdateLatestStable = QAction(MondoMainWindow)
        self.actionUpdateLatestStable.setObjectName(u"actionUpdateLatestStable")
        self.actionUpdateSpecificBranch = QAction(MondoMainWindow)
        self.actionUpdateSpecificBranch.setObjectName(u"actionUpdateSpecificBranch")
        self.actionUpdateSpecificCommit = QAction(MondoMainWindow)
        self.actionUpdateSpecificCommit.setObjectName(u"actionUpdateSpecificCommit")
        self.actionUpdateCurrentBranch = QAction(MondoMainWindow)
        self.actionUpdateCurrentBranch.setObjectName(u"actionUpdateCurrentBranch")
        self.actionPreferences = QAction(MondoMainWindow)
        self.actionPreferences.setObjectName(u"actionPreferences")
        self.actionShowLogs = QAction(MondoMainWindow)
        self.actionShowLogs.setObjectName(u"actionShowLogs")
        self.centralwidget = QWidget(MondoMainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_5 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.groupBox = QGroupBox(self.centralwidget)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.psdButton = QPushButton(self.groupBox)
        self.psdButton.setObjectName(u"psdButton")
        self.psdButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout.addWidget(self.psdButton)

        self.singleChannelPsdButton = QPushButton(self.groupBox)
        self.singleChannelPsdButton.setObjectName(u"singleChannelPsdButton")

        self.verticalLayout.addWidget(self.singleChannelPsdButton)

        self.singleSlicePsdButton = QPushButton(self.groupBox)
        self.singleSlicePsdButton.setObjectName(u"singleSlicePsdButton")

        self.verticalLayout.addWidget(self.singleSlicePsdButton)

        self.overlaidPsdButton = QPushButton(self.groupBox)
        self.overlaidPsdButton.setObjectName(u"overlaidPsdButton")
        self.overlaidPsdButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout.addWidget(self.overlaidPsdButton)

        self.verticalSpacer = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.gridLayout.addWidget(self.groupBox, 0, 2, 1, 1)

        self.groupBox_4 = QGroupBox(self.centralwidget)
        self.groupBox_4.setObjectName(u"groupBox_4")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox_4)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.csvButton = QPushButton(self.groupBox_4)
        self.csvButton.setObjectName(u"csvButton")
        self.csvButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_4.addWidget(self.csvButton)

        self.csvDownsampledButton = QPushButton(self.groupBox_4)
        self.csvDownsampledButton.setObjectName(u"csvDownsampledButton")
        self.csvDownsampledButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_4.addWidget(self.csvDownsampledButton)

        self.verticalSpacer_4 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer_4)


        self.gridLayout.addWidget(self.groupBox_4, 1, 0, 1, 1)

        self.groupBox_2 = QGroupBox(self.centralwidget)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout_2 = QVBoxLayout(self.groupBox_2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.spectrogramButton = QPushButton(self.groupBox_2)
        self.spectrogramButton.setObjectName(u"spectrogramButton")
        self.spectrogramButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_2.addWidget(self.spectrogramButton)

        self.verticalSpacer_2 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer_2)


        self.gridLayout.addWidget(self.groupBox_2, 0, 1, 1, 1)

        self.groupBox_3 = QGroupBox(self.centralwidget)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox_3)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.timeButton = QPushButton(self.groupBox_3)
        self.timeButton.setObjectName(u"timeButton")
        self.timeButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_3.addWidget(self.timeButton)

        self.synchronousTimeButton = QPushButton(self.groupBox_3)
        self.synchronousTimeButton.setObjectName(u"synchronousTimeButton")
        self.synchronousTimeButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_3.addWidget(self.synchronousTimeButton)

        self.overlaidTimeButton = QPushButton(self.groupBox_3)
        self.overlaidTimeButton.setObjectName(u"overlaidTimeButton")
        self.overlaidTimeButton.setMinimumSize(QSize(175, 0))

        self.verticalLayout_3.addWidget(self.overlaidTimeButton)

        self.verticalSpacer_3 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_3.addItem(self.verticalSpacer_3)


        self.gridLayout.addWidget(self.groupBox_3, 0, 0, 1, 1)

        self.groupBox_5 = QGroupBox(self.centralwidget)
        self.groupBox_5.setObjectName(u"groupBox_5")
        self.verticalLayout_6 = QVBoxLayout(self.groupBox_5)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.fileInformationButton = QPushButton(self.groupBox_5)
        self.fileInformationButton.setObjectName(u"fileInformationButton")

        self.verticalLayout_6.addWidget(self.fileInformationButton)

        self.viewSettingsButton = QPushButton(self.groupBox_5)
        self.viewSettingsButton.setObjectName(u"viewSettingsButton")

        self.verticalLayout_6.addWidget(self.viewSettingsButton)

        self.rawExportButton = QPushButton(self.groupBox_5)
        self.rawExportButton.setObjectName(u"rawExportButton")

        self.verticalLayout_6.addWidget(self.rawExportButton)

        self.splitFileButton = QPushButton(self.groupBox_5)
        self.splitFileButton.setObjectName(u"splitFileButton")

        self.verticalLayout_6.addWidget(self.splitFileButton)

        self.verticalSpacer_6 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_6.addItem(self.verticalSpacer_6)


        self.gridLayout.addWidget(self.groupBox_5, 1, 1, 1, 1)


        self.verticalLayout_5.addLayout(self.gridLayout)

        self.verticalSpacer_5 = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_5.addItem(self.verticalSpacer_5)

        self.verticalLayout_5.setStretch(1, 1)
        MondoMainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MondoMainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 617, 22))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        self.menuCheckForUpdates = QMenu(self.menuHelp)
        self.menuCheckForUpdates.setObjectName(u"menuCheckForUpdates")
        self.menuSettings = QMenu(self.menubar)
        self.menuSettings.setObjectName(u"menuSettings")
        MondoMainWindow.setMenuBar(self.menubar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuSettings.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionExit)
        self.menuHelp.addAction(self.menuCheckForUpdates.menuAction())
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionShowLogs)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionAboutLibraries)
        self.menuCheckForUpdates.addAction(self.actionUpdateLatestStable)
        self.menuCheckForUpdates.addAction(self.actionUpdateCurrentBranch)
        self.menuCheckForUpdates.addAction(self.actionUpdateSpecificBranch)
        self.menuCheckForUpdates.addAction(self.actionUpdateSpecificCommit)
        self.menuSettings.addAction(self.actionPreferences)

        self.retranslateUi(MondoMainWindow)
        self.actionExit.triggered.connect(MondoMainWindow.close)

        QMetaObject.connectSlotsByName(MondoMainWindow)
    # setupUi

    def retranslateUi(self, MondoMainWindow):
        MondoMainWindow.setWindowTitle(QCoreApplication.translate("MondoMainWindow", u"Mondo", None))
        self.actionAbout.setText(QCoreApplication.translate("MondoMainWindow", u"About Mondo...", None))
#if QT_CONFIG(tooltip)
        self.actionAbout.setToolTip(QCoreApplication.translate("MondoMainWindow", u"About Mondo...", None))
#endif // QT_CONFIG(tooltip)
        self.actionExit.setText(QCoreApplication.translate("MondoMainWindow", u"Exit", None))
        self.actionAboutLibraries.setText(QCoreApplication.translate("MondoMainWindow", u"About Libraries...", None))
#if QT_CONFIG(tooltip)
        self.actionAboutLibraries.setToolTip(QCoreApplication.translate("MondoMainWindow", u"About Libraries...", None))
#endif // QT_CONFIG(tooltip)
        self.actionUpdateLatestStable.setText(QCoreApplication.translate("MondoMainWindow", u"Latest Stable", None))
        self.actionUpdateSpecificBranch.setText(QCoreApplication.translate("MondoMainWindow", u"Specific Branch...", None))
        self.actionUpdateSpecificCommit.setText(QCoreApplication.translate("MondoMainWindow", u"Specific Commit..", None))
        self.actionUpdateCurrentBranch.setText(QCoreApplication.translate("MondoMainWindow", u"Latest branch", None))
        self.actionPreferences.setText(QCoreApplication.translate("MondoMainWindow", u"Preferences...", None))
#if QT_CONFIG(tooltip)
        self.actionPreferences.setToolTip(QCoreApplication.translate("MondoMainWindow", u"Preferences...", None))
#endif // QT_CONFIG(tooltip)
        self.actionShowLogs.setText(QCoreApplication.translate("MondoMainWindow", u"Show Logs on Disk...", None))
        self.groupBox.setTitle(QCoreApplication.translate("MondoMainWindow", u"Frequency Analysis", None))
        self.psdButton.setText(QCoreApplication.translate("MondoMainWindow", u"Power Spectral Density (PSD)", None))
#if QT_CONFIG(tooltip)
        self.singleChannelPsdButton.setToolTip(QCoreApplication.translate("MondoMainWindow", u"Overlaid PSD streamlined for plotting a single channel across multiple time slices", None))
#endif // QT_CONFIG(tooltip)
        self.singleChannelPsdButton.setText(QCoreApplication.translate("MondoMainWindow", u"Single Channel Overlaid PSD", None))
#if QT_CONFIG(tooltip)
        self.singleSlicePsdButton.setToolTip(QCoreApplication.translate("MondoMainWindow", u"Overlaid PSD streamlined for plotting multiple channels in a single time slice", None))
#endif // QT_CONFIG(tooltip)
        self.singleSlicePsdButton.setText(QCoreApplication.translate("MondoMainWindow", u"Single Time Slice Overlaid PSD", None))
#if QT_CONFIG(tooltip)
        self.overlaidPsdButton.setToolTip(QCoreApplication.translate("MondoMainWindow", u"Ovelaid PSD for plotting arbitrary channels and time slices ", None))
#endif // QT_CONFIG(tooltip)
        self.overlaidPsdButton.setText(QCoreApplication.translate("MondoMainWindow", u"Overlaid PSD", None))
        self.groupBox_4.setTitle(QCoreApplication.translate("MondoMainWindow", u"Raw Data", None))
        self.csvButton.setText(QCoreApplication.translate("MondoMainWindow", u"CSV Output", None))
        self.csvDownsampledButton.setText(QCoreApplication.translate("MondoMainWindow", u"CSV Downsampled Output", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("MondoMainWindow", u"Time/Frequency Analysis", None))
        self.spectrogramButton.setText(QCoreApplication.translate("MondoMainWindow", u"Spectrogram", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("MondoMainWindow", u"Time Analysis", None))
        self.timeButton.setText(QCoreApplication.translate("MondoMainWindow", u"Time Data", None))
        self.synchronousTimeButton.setText(QCoreApplication.translate("MondoMainWindow", u"Synchronous Time Data", None))
        self.overlaidTimeButton.setText(QCoreApplication.translate("MondoMainWindow", u"Overlaid Time Data", None))
        self.groupBox_5.setTitle(QCoreApplication.translate("MondoMainWindow", u"File Manipulation", None))
        self.fileInformationButton.setText(QCoreApplication.translate("MondoMainWindow", u"Device Information", None))
        self.viewSettingsButton.setText(QCoreApplication.translate("MondoMainWindow", u"View Device Settings", None))
        self.rawExportButton.setText(QCoreApplication.translate("MondoMainWindow", u"Advanced Raw Export", None))
        self.splitFileButton.setText(QCoreApplication.translate("MondoMainWindow", u"Split File", None))
        self.menuFile.setTitle(QCoreApplication.translate("MondoMainWindow", u"File", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MondoMainWindow", u"Help", None))
        self.menuCheckForUpdates.setTitle(QCoreApplication.translate("MondoMainWindow", u"Check for Updates", None))
        self.menuSettings.setTitle(QCoreApplication.translate("MondoMainWindow", u"Settings", None))
    # retranslateUi




# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import math
from typing import Literal

from matplotlib.backend_bases import MouseButton
from matplotlib.figure import Figure
import matplotlib.mlab as mlab
import numpy
from numpy.typing import NDArray
from PySide6 import QtGui, QtWidgets

from . import util
from .. import export_csv
from .. import export_script

logger = logging.getLogger(__name__)


def get_integrated_power(start_freq: float, end_freq: float,
                         freqs: NDArray[numpy.float64],
                         pxx: NDArray[numpy.float64]) -> float:
    start_index: int = max(0, numpy.searchsorted(
        freqs, start_freq, side="right").item() - 1)
    end_index: int = min(len(freqs), numpy.searchsorted(
        freqs, end_freq, side="left").item() + 1)

    dfreq = freqs[1] - freqs[0]
    integrated = numpy.sum(pxx[start_index:end_index]) * dfreq
    return integrated.item()


def add_amplitude_button(fig: Figure, ax, data: list[
        tuple[NDArray[numpy.float64], NDArray[numpy.float64], str,
              util.UnitInfo]]) -> None:
    rect = None  # used in the callback

    toolbar: QtWidgets.QToolBar = fig.canvas.toolbar  # type: ignore
    actionText = QtWidgets.QApplication.translate("AmplitudeAction",
                                                  "Amplitude Integration")
    action = QtGui.QAction(actionText, toolbar)
    action.setIcon(QtGui.QIcon.fromTheme("measuring_cup"))

    def measure_sine_amplitude() -> None:
        QtWidgets.QMessageBox.information(
            toolbar, "Select Integration Range",
            "Select integration range.\nUndo = middle click / backspace")

        # grab the points from the user
        points = fig.ginput(n=2, timeout=0, mouse_add=MouseButton.LEFT,
                            mouse_stop=None,  # type: ignore
                            mouse_pop=MouseButton.MIDDLE)
        if len(points) != 2:
            return  # user cancelled
        start_freq = min(p[0] for p in points)
        end_freq = max(p[0] for p in points)

        # show a box with the freq range
        nonlocal rect
        if rect:
            rect.remove()
            rect = None
        rect = ax.axvspan(start_freq, end_freq, facecolor='g', alpha=0.5)
        fig.canvas.draw()

        value_strings: list[str] = []

        for freqs, pxx, name, unit in data:
            power = get_integrated_power(start_freq, end_freq, freqs, pxx)
            amplitude = power ** 0.5

            if name:
                value_strings.append(f"{name}: {amplitude} {unit.utf8}")
            else:
                value_strings.append(f"{amplitude} {unit.utf8}")

        if len(value_strings) == 1:
            header = "Amplitude (RMS):"
        else:
            header = "Amplitudes (RMS):"

        value_strings.insert(0, header)
        text = "\n".join(value_strings)

        QtWidgets.QMessageBox.information(
            toolbar, "Amplitudes", text)

    action.triggered.connect(measure_sine_amplitude)

    toolbar.addAction(action)


def _stride_windows(x: NDArray[numpy.float64], n: int,
                    noverlap: int) -> NDArray[numpy.float64]:
    # JSN this was pulled from matplotlib.mlab._stride_windows, and streamlined
    return numpy.lib.stride_tricks.sliding_window_view(
        x, n, axis=0)[::n - noverlap].T


def do_psd(sequence: list[util.Chunk], subchannel_index: int, NFFT: int,
           Fs: float, detrend: Literal['mean', 'linear', 'none'],
           window: NDArray[numpy.float64],
           noverlap: int) -> tuple[NDArray[numpy.float64],
                                   NDArray[numpy.float64]]:
    # JSN NOTE: much of this function was taken from matplotlib.mlab's
    # internal _spectral_helper(). It has been streamlined by removing
    # unneeded features, and changed to support chunks of data

    freqs = numpy.fft.rfftfreq(NFFT, 1 / Fs)

    windows = 0
    sums = numpy.zeros(freqs.shape)

    for _time, data, _start, _end in sequence:
        x = numpy.asarray(data[:, subchannel_index])

        if len(x) < NFFT:
            continue

        result = _stride_windows(x, NFFT, noverlap)
        result = mlab.detrend(result, detrend, axis=0)
        result = numpy.asarray(result) * window.reshape((-1, 1))
        result = numpy.fft.rfft(result, n=NFFT, axis=0)

        result = (numpy.conjugate(result) * result).real

        # Scale the spectrum by the norm of the window to compensate for
        # windowing loss; see Bendat & Piersol Sec 11.5.2.
        result /= (numpy.abs(window) ** 2).sum()

        windows += result.shape[1]
        result = result.sum(axis=1)
        sums += result

        del result

    sums[1:] *= 2.0  # because it's one sided

    sums /= Fs  # get the spectral density

    sums /= windows

    return sums, freqs


def psd_analysis(parent: QtWidgets.QWidget) -> None:
    ret = util.load_batch(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_index = util.choose_channel(header, parent)
    if channel_index is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, [channel_index], parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    subchannel_index = util.choose_subchannel(
        batch_info[channel_index], allow_all=False, parent=parent)
    if subchannel_index is None:
        return  # user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret_lost_packets = util.warn_about_lost_packets(batch_info, parent)
    if not ret_lost_packets:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(
        batch_info, parent, chunk=True, raw_units=True)

    sequence = sequence_info[channel_index]
    unit = unit_info[channel_index]

    sampling_rate = None
    for time, _data, _start, _end in sequence:
        if len(time) > 2:
            sampling_rate = 1.0 / (time[1] - time[0])
            break

    if sampling_rate is None:
        message = "Could not determine sampling rate!"
        logger.error(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    psd_options = util.get_psd_options([sequence], sampling_rate, parent)
    if psd_options is None:
        return  # user cancelled

    pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
    pxx_dB = 10 * numpy.log10(pxx)

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    ax.plot(freqs, pxx_dB)
    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        step = 1
    else:
        step = logi * 10
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)  # type: ignore
    ax.get_yaxis().get_major_formatter().set_useOffset(False)  # type: ignore

    export_csv.add_export_csv_action(fig, labels, [(freqs, pxx_dB, None)])

    add_amplitude_button(object.__getattribute__(fig, '_wrapped'),
                         object.__getattribute__(ax, '_wrapped'),
                         [(freqs, pxx, "", unit)])

    fig.show()


def single_channel_psd_analysis(parent: QtWidgets.QWidget) -> None:
    channel_index = None  # ask for the channel on the first iteration
    subchannel_index = None  # ask for the subchannel on the first iteration

    sampling_rate = None
    sequences: list[list[util.Chunk]] = []
    units: list[util.UnitInfo] = []
    names = []
    files = set()

    while True:
        # load a batch of files
        ret = util.load_batch(parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # warn the user if any files have been loaded before
        for filename, _dt in file_infos:
            if filename in files:
                message = ("File {} already loaded in another batch!\n"
                           "Proceed with caution!").format(filename)
                logger.warning(message)
                QtWidgets.QMessageBox.warning(parent, "Warning", message)
            files.add(filename)

        # if it's the first time through, choose a channel
        if channel_index is None:
            channel_index = util.choose_channel(header, parent)
            if channel_index is None:
                return  # error message displayed, or user cancelled

        # make sure this batch has a channel of the correct index
        if channel_index >= len(header['channels']):
            message = "Channels inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        batch_info = util.decode_batch(
            file_infos, header, [channel_index], parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        if subchannel_index is None:
            subchannel_index = util.choose_subchannel(
                batch_info[channel_index], allow_all=False, parent=parent)
            if subchannel_index is None:
                return  # user cancelled

        subchannels = batch_info[channel_index].channel_decoder.subchannels
        if subchannel_index >= subchannels:
            message = "Channels inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        batch_info = util.get_datetime_subset(batch_info, parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret_lost_packets = util.warn_about_lost_packets(batch_info, parent)
        if not ret_lost_packets:
            return  # user cancelled

        sequence_info, unit_info = util.sequence_data(
            batch_info, parent, chunk=True, raw_units=True)

        sequence = sequence_info[channel_index]

        sequences.append(sequence)
        units.append(unit_info[channel_index])

        batch_sampling_rate = None
        for time, _data, _start, _end in sequence:
            if len(time) > 2:
                batch_sampling_rate = 1.0 / (time[1] - time[0])
                break

        if batch_sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        if sampling_rate is None:
            sampling_rate = batch_sampling_rate
        elif sampling_rate != batch_sampling_rate:
            message = "Sampling rates inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret_dialog = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.StandardButton.Yes |
                     QtWidgets.QMessageBox.StandardButton.No),
            defaultButton=QtWidgets.QMessageBox.StandardButton.No)
        if ret_dialog != QtWidgets.QMessageBox.StandardButton.Yes:
            break

    psd_options = util.get_psd_options(sequences, sampling_rate, parent)
    if psd_options is None:
        return  # user cancelled

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []
    amplitude_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                               str, util.UnitInfo]] = []

    for i, sequence in enumerate(sequences):
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=names[i])
        export_data.append((freqs, pxx_dB, names[i]))
        amplitude_data.append((freqs, pxx, names[i], units[i]))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        step = 1
    else:
        step = logi * 10
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)  # type: ignore
    ax.get_yaxis().get_major_formatter().set_useOffset(False)  # type: ignore

    export_csv.add_export_csv_action(fig, labels, export_data)

    add_amplitude_button(object.__getattribute__(fig, '_wrapped'),
                         object.__getattribute__(ax, '_wrapped'),
                         amplitude_data)

    fig.show()


def single_slice_psd_analysis(parent: QtWidgets.QWidget) -> None:
    ret = util.load_batch(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    subchannels_list = util.choose_subchannels(header, parent)
    if not subchannels_list:
        return  # error message displayed, or user cancelled

    channel_indexes = {t[0] for t in subchannels_list}

    batch_info = util.decode_batch(file_infos, header, channel_indexes, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret_lost_packets = util.warn_about_lost_packets(batch_info, parent)
    if not ret_lost_packets:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(
        batch_info, parent, chunk=True, raw_units=True)

    sampling_rates = {}  # key: channel, value: sampling rate
    for channel_index in channel_indexes:
        sampling_rate = None
        for time, _data, _start, _end in sequence_info[channel_index]:
            if len(time) > 2:
                sampling_rate = 1.0 / (time[1] - time[0])
                break

        if sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        sampling_rates[channel_index] = sampling_rate

    # create a list of similar channels, which can share psd options
    channel_signatures = {}
    for channel_index in channel_indexes:
        sequence = sequence_info[channel_index]
        sequence_signature = tuple(len(x[0]) for x in sequence)
        channel_signatures[channel_index] = (sampling_rates[channel_index],
                                             sequence_signature)

    remaining_channels = set(channel_indexes)
    sections: list[tuple[str, list[list[util.Chunk]], float]] = []
    channel_groups = []
    for channel_index in sorted(channel_indexes):
        if channel_index in remaining_channels:
            remaining_channels.discard(channel_index)
        else:
            continue

        sig = channel_signatures[channel_index]
        similar_channels = [channel_index]
        for other_channel_index in sorted(remaining_channels):
            if sig == channel_signatures[other_channel_index]:
                similar_channels.append(other_channel_index)
                remaining_channels.discard(other_channel_index)

        group_name = ", ".join((batch_info[i].channel_decoder.channel_name
                                for i in similar_channels))

        # the sequence and rate of all the similar channels are the same
        sequence = sequence_info[similar_channels[0]]
        sampling_rate = sampling_rates[similar_channels[0]]

        sections.append((group_name, [sequence], sampling_rate))
        channel_groups.append(similar_channels)

    all_psd_options = util.get_multiple_psd_options(sections, parent)
    if all_psd_options is None:
        return  # user cancelled

    # unpack psd options, per channel
    channel_psd_options = {}
    for i, psd_options in enumerate(all_psd_options):
        for channel_index in channel_groups[i]:
            channel_psd_options[channel_index] = psd_options

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []
    amplitude_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                               str, util.UnitInfo]] = []

    for channel_index, subchannel_index in subchannels_list:
        channel_decoder = batch_info[channel_index].channel_decoder
        if len(channel_decoder.subchannel_names) > 1:
            name = channel_decoder.subchannel_names[subchannel_index]
        else:
            name = channel_decoder.channel_name

        sequence = sequence_info[channel_index]
        psd_options = channel_psd_options[channel_index]
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=name)
        export_data.append((freqs, pxx_dB, name))
        amplitude_data.append((freqs, pxx, name, unit_info[channel_index]))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        step = 1
    else:
        step = logi * 10
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)  # type: ignore
    ax.get_yaxis().get_major_formatter().set_useOffset(False)  # type: ignore

    export_csv.add_export_csv_action(fig, labels, export_data)

    add_amplitude_button(object.__getattribute__(fig, '_wrapped'),
                         object.__getattribute__(ax, '_wrapped'),
                         amplitude_data)

    fig.show()


def overlaid_psd_analysis(parent: QtWidgets.QWidget) -> None:
    sequences: list[tuple[list[util.Chunk], util.UnitInfo]] = []
    names = []
    sampling_rates = []

    while True:
        # load a batch of files
        ret = util.load_batch(parent=parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # if it's the first time through, choose a channel
        channel_index = util.choose_channel(header, parent=parent)
        if channel_index is None:
            return  # error message displayed, or user cancelled

        batch_info = util.decode_batch(file_infos, header, [channel_index],
                                       parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        subchannel_index = util.choose_subchannel(
            batch_info[channel_index], allow_all=False, parent=parent)
        if subchannel_index is None:
            return  # user cancelled

        batch_info = util.get_datetime_subset(batch_info, parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret_lost_packets = util.warn_about_lost_packets(batch_info, parent)
        if not ret_lost_packets:
            return  # user cancelled

        sequence_info, unit_info = util.sequence_data(
            batch_info, parent, chunk=True, raw_units=True)

        sequence = sequence_info[channel_index]
        unit = unit_info[channel_index]

        sequences.append((sequence, unit))

        sampling_rate = None
        for time, _data, _start, _end in sequence:
            if len(time) > 2:
                sampling_rate = 1.0 / (time[1] - time[0])
                break

        if sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        sampling_rates.append(sampling_rate)

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret_dialog = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.StandardButton.Yes |
                     QtWidgets.QMessageBox.StandardButton.No),
            defaultButton=QtWidgets.QMessageBox.StandardButton.No)
        if ret_dialog != QtWidgets.QMessageBox.StandardButton.Yes:
            break

    sections: list[tuple[str, list[list[util.Chunk]], float]] = []
    for i, (sequence, _unit) in enumerate(sequences):
        sections.append((names[i], [sequence], sampling_rates[i]))

    all_psd_options = util.get_multiple_psd_options(sections, parent)
    if all_psd_options is None:
        return  # user cancelled

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []
    amplitude_data: list[tuple[NDArray[numpy.float64], NDArray[numpy.float64],
                               str, util.UnitInfo]] = []

    for i, (sequence, unit) in enumerate(sequences):
        psd_options = all_psd_options[i]
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=names[i])
        export_data.append((freqs, pxx_dB, names[i]))
        amplitude_data.append((freqs, pxx, names[i], unit))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        step = 1
    else:
        step = logi * 10
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)  # type: ignore
    ax.get_yaxis().get_major_formatter().set_useOffset(False)  # type: ignore

    export_csv.add_export_csv_action(fig, labels, export_data)

    add_amplitude_button(object.__getattribute__(fig, '_wrapped'),
                         object.__getattribute__(ax, '_wrapped'),
                         amplitude_data)

    fig.show()
